@@ -20,23 +20,7 @@ See https://sidsite.com/posts/autodiff/
 from collections import defaultdict
 import math
 import numpy
-
-
-class ArrayLibraryProxy:
-    """Enable switching between NumPy and CuPy dynamically.
-
-    E.g.
-    >> import smallpebble as sp
-    >> import cupy
-    >> sp.array_library = cupy
-    """
-
-    def __getattribute__(self, name):
-        return getattr(array_library, name)
-
-
-array_library = numpy
-np = ArrayLibraryProxy()
+import smallpebble.array_library as np
 
 
 # ----------------
@@ -235,13 +219,22 @@ def matrix_transpose(a):
 
 def maxax(a, axis):
     "Reduce an axis, `axis`, to its max value."
-    max_idx = np.argmax(a.array, axis)
-    max_idx = np.expand_dims(max_idx, axis)
-    value = np.take_along_axis(a.array, max_idx, axis)
+    # Note, implementation now more complicated because CuPy doesn't have put_along_axis.
+    axis = axis if axis >= 0 else a.ndim + axis
+    value = np.swapaxes(a.array, axis, -1)
+    value = value.reshape([-1, value.shape[-1]])
+    flatshape = value.shape
+    idx = np.argmax(value, axis=-1)
+    value = np.take_along_axis(value, idx[..., np.newaxis], -1)
+    value = value.reshape(tuple(1 if i == axis else v for i, v in enumerate(a.shape)))
 
     def multiply_by_locgrad(path_value):
-        result = np.zeros(a.array.shape)
-        np.put_along_axis(result, max_idx, np.array(1, a.array.dtype), axis)
+        result = np.zeros(flatshape)
+        result[np.arange(result.shape[0]), idx] = 1
+        swapped_shape = list(a.shape)
+        swapped_shape[axis], swapped_shape[-1] = swapped_shape[-1], swapped_shape[axis]
+        result = result.reshape(swapped_shape)
+        result = np.swapaxes(result, axis, -1)
         return path_value * result
 
     local_gradients = ((a, multiply_by_locgrad),)
@@ -607,22 +600,24 @@ def broadcastinfo(a_shape, b_shape):
     a_repeatdims = (a_shape_ == 1) & (b_shape_ > 1)  # the repeated dims
     a_repeatdims[:add_ndims_to_a] = True  # the added dims
     a_repeatdims = np.where(a_repeatdims == True)[0]  # indices of axes where True
+    a_repeatdims = [int(i) for i in a_repeatdims]
 
     b_repeatdims = (b_shape_ == 1) & (a_shape_ > 1)
     b_repeatdims[:add_ndims_to_b] = True
     b_repeatdims = np.where(b_repeatdims == True)[0]
+    b_repeatdims = [int(i) for i in b_repeatdims]
 
     return tuple(a_repeatdims), tuple(b_repeatdims)
 
 
 def np_add_at(a, indices, b):
     "Apply either numpy.add.at or cupy.scatter_add, depending on which library is used."
-    if array_library.__name__ == "numpy":
-        array_library.add.at(a, indices, b)
-    elif array_library.__name__ == "cupy":
-        array_library.scatter_add(a, indices, b)
+    if np.library.__name__ == "numpy":
+        np.add.at(a, indices, b)
+    elif np.library.__name__ == "cupy":
+        np.scatter_add(a, indices, b)
     else:
-        raise ValueError("Expected array_library.__name__ to be `numpy` or `cupy`.")
+        raise ValueError("Expected np.library.__name__ to be `numpy` or `cupy`.")
 
 
 def np_strided_sliding_view(x, window_shape: tuple, strides: tuple):
